def return_errors():

    {
        "errors": [
            {
                "code": "<error identifier>",
                "message": "<message describing condition>",
                "detail": "<unstructured>",
            },
            ...,
        ]
    }
